import os
import numpy as np

from bmtk.builder import NetworkBuilder
from bmtk.builder.auxi.node_params import positions_columinar, xiter_random


# Use seed 0 to allows replicable networks everytime it is rebuilt. Removing or changing seed value will build
# a network with similar architect but different connection probabilities.
np.random.seed(0)

build_recurrent_edges = True

print('Building internal network')
# List of non-virtual cell models
cell_models = [
    {
        'model_name': 'Scnn1a', 'ei': 'e',
        'model_template': 'nest:iaf_psc_alpha',
        'dynamics_params': '472363762_point.json'
    },
    {
        'model_name': 'Rorb', 'ei': 'e',
        'model_template': 'nest:iaf_psc_alpha',
        'dynamics_params': '473863510_point.json'
    },
    {
        'model_name': 'Nr5a1', 'ei': 'e',
        'model_template': 'nest:iaf_psc_alpha',
        'dynamics_params': '473863035_point.json'
    },
    {
        'model_name': 'PV1', 'ei': 'i',
        'model_template': 'nest:iaf_psc_alpha',
        'dynamics_params': '472912177_point.json'
    },
    {
        'model_name': 'PV2', 'ei': 'i',
        'model_template': 'nest:iaf_psc_alpha',
        'dynamics_params': '473862421_point.json'
    }
]

# Build a network of 300 biophysical cells to simulate
internal = NetworkBuilder("internal")
for i, model_props in enumerate(cell_models):
    n_cells = 80 if model_props['ei'] == 'e' else 30  # 80% excitatory, 20% inhib

    # Randomly get positions uniformly distributed in a column
    positions = positions_columinar(N=n_cells, center=[0, 10.0, 0], max_radius=50.0, height=200.0)

    internal.add_nodes(N=n_cells,
                       x=positions[:, 0], y=positions[:, 1], z=positions[:, 2],
                       rotation_angle_yaxis=xiter_random(N=n_cells, min_x=0.0, max_x=2 * np.pi),  # randomly rotate y axis
                       model_type='point_process',
                       **model_props)



if build_recurrent_edges:
    def n_connections(src, trg, prob=0.5, min_syns=2, max_syns=7):
        return 0 if np.random.uniform() > prob else np.random.randint(min_syns, max_syns)

    # exc --> exc connections
    cm = internal.add_edges(source={'ei': 'e'}, target={'ei': 'e'},
                            connection_rule=lambda s, t: np.random.binomial(1, 0.2),
                            dynamics_params='ExcToExc.json',
                            model_template='static_synapse',
                            # syn_weight=2.5,
                            delay=2.0)
    cm.add_properties('syn_weight', rule=2.5, dtypes=np.float)

    # exc --> inh connections
    cm = internal.add_edges(source={'ei': 'e'}, target={'ei': 'i'},
                            connection_rule=lambda s, t: np.random.binomial(1, 0.5),
                            dynamics_params='ExcToInh.json',
                            model_template='static_synapse',
                            # syn_weight=7.0,
                            delay=2.0)
    cm.add_properties('syn_weight', rule=7.0, dtypes=np.float)

    # inh --> exc connections
    cm = internal.add_edges(source={'ei': 'i'}, target={'ei': 'e'},
                            connection_rule=lambda s, t: np.random.binomial(1, 0.5),
                            #connection_rule=lambda *_: np.random.randint(0, 4),
                            dynamics_params='InhToExc.json',
                            model_template='static_synapse',
                            # syn_weight=-7.5,
                            delay=2.0)
    cm.add_properties('syn_weight', rule=-7.5, dtypes=np.float)

    # inh --> inh connections
    cm = internal.add_edges(source={'ei': 'i'}, target={'ei': 'i'},
                            connection_rule=lambda s, t: np.random.binomial(1, 0.5),
                            dynamics_params='InhToInh.json',
                            model_template='static_synapse',
                            # syn_weight=-3.0,
                            delay=2.0)
    cm.add_properties('syn_weight', rule=-3.0, dtypes=np.float)


internal.build()

print('Saving internal')
internal.save(output_dir='network')


print('Building external connections')
external = NetworkBuilder("external")
external.add_nodes(N=100, model_type='virtual', ei='e')
cm = external.add_edges(target=internal.nodes(ei='e'), source=external.nodes(),
                        connection_rule=lambda *_: np.random.binomial(1, .7),
                        dynamics_params='ExcToExc.json',
                        model_template='static_synapse')
cm.add_properties('syn_weight', rule=50.0, dtypes=np.float)


cm = external.add_edges(target=internal.nodes(ei='i'), source=external.nodes(),
                        connection_rule=lambda *_: np.random.binomial(1, .7),
                        dynamics_params='ExcToInh.json',
                        model_template='static_synapse')
cm.add_properties('syn_weight', rule=65.0, dtypes=np.float)


external.build()

print('Saving external')
external.save(output_dir='network')


